import RNG from './RNG'

const SCREEN_WIDTH = 64;
const SCREEN_HEIGHT = 32;

const MEMORY_SIZE = 4096;
const STACK_SIZE = 16;

const REGISTERS = 16;

class CPU {
	pc: number; // 16b
	indexReg: number; // 16b
	registers: Uint8Array;

	delayTimer: number; // 8b
	soundTimer: number; // 8b

	stackPointer: number; // 16b
	stack: Array<number>;

	memory: Uint8Array;
	displayMemory: Uint8Array;
	
	rng: RNG;

	constructor() {
		this.memory = new Uint8Array(MEMORY_SIZE);
		this.displayMemory = new Uint8Array(SCREEN_WIDTH*SCREEN_HEIGHT);
		this.pc = 0x200;
		this.indexReg = 0;
		this.stackPointer = 0;
		this.stack = new Array<number>(STACK_SIZE);
		this.delayTimer = 0;
		this.soundTimer = 0;
		this.registers = new Uint8Array(REGISTERS);
		
		this.rng = new RNG(0);

		this.clearMemory();
		this.clearDisplay();
		this.clearRegisters();

		// 0x000-0x1FF - Chip 8 interpreter
		// 0x050-0x0A0 - 4x5 pixel font set (0-F)
		// 0x200-0xFFF - Program ROM and work RAM
		this.setupFont();
	}

	clearRegisters() {
		this.registers.fill(0);
		this.pc = 0x200; // Start of the program on normal programs
		this.indexReg = 0;
		this.stackPointer = 0;
		this.delayTimer = 0;
		this.soundTimer = 0;

		this.rng = new RNG(0);
	}

	clearMemory() {
		this.memory.fill(0);
	}

	clearDisplay() {
		this.displayMemory.fill(0);
	}


	private setupFont() {
		// Using address from 050 - 09F ??
		let addressFont = 0x050;
		
		// Every character has 5 Bytes
		const characters = [
			0xF0, 0x90, 0x90, 0x90, 0xF0, // 0
			0x20, 0x60, 0x20, 0x20, 0x70, // 1
			0xF0, 0x10, 0xF0, 0x80, 0xF0, // 2
			0xF0, 0x10, 0xF0, 0x10, 0xF0, // 3
			0x90, 0x90, 0xF0, 0x10, 0x10, // 4
			0xF0, 0x80, 0xF0, 0x10, 0xF0, // 5
			0xF0, 0x80, 0xF0, 0x90, 0xF0, // 6
			0xF0, 0x10, 0x20, 0x40, 0x40, // 7
			0xF0, 0x90, 0xF0, 0x90, 0xF0, // 8
			0xF0, 0x90, 0xF0, 0x10, 0xF0, // 9
			0xF0, 0x90, 0xF0, 0x90, 0x90, // A
			0xE0, 0x90, 0xE0, 0x90, 0xE0, // B
			0xF0, 0x80, 0x80, 0x80, 0xF0, // C
			0xE0, 0x90, 0x90, 0x90, 0xE0, // D
			0xF0, 0x80, 0xF0, 0x80, 0xF0, // E
			0xF0, 0x80, 0xF0, 0x80, 0x80  // F
		];

		this.memory.set(characters, addressFont);
	}

	private incrementPC() {
		this.pc += 2;
	}
	
	private cycle() {
		// Fetch opcode
		let firstOpcode: number = this.memory[this.pc];
		let secondOpcode: number = this.memory[this.pc + 1];
		let opcode: number = firstOpcode << 8 | secondOpcode;
		
		// Decode opcode
		// X000
		let firstNibble = firstOpcode >> 4;
		
		switch(firstNibble) {
			case 0x0: 
				if(opcode == 0x00E0) {
					// Clear the screen
					this.clearDisplay();
				}else if(opcode == 0x00EE) {
					// Returns from a subroutine
					this.stackPointer--;
					this.pc = this.stack[this.pc];
				}
				this.incrementPC();
				break;
			case 0x1:
				// goto NNN;
				let threeNib = opcode & 0x0FFF;
				this.pc = threeNib;
				break;
			case 0x2:
				// Calls subroutine at NNN
				this.stack[this.stackPointer] = this.pc; // Store PC at stack
				this.stackPointer++;	// Increment it
				this.pc = opcode & 0x0FFF; // Go to that address to execute code
				break;
			case 0x3: {
				// 3XNN
				// Skips the next instruction if VX equals NN
				let secondNibble = firstOpcode & 0xF; //0X00
				let val = opcode & 0x00FF; //0x00NN
				if(this.registers[secondNibble] == val){
					this.incrementPC();
				}
				this.incrementPC();
				}
				break;
			case 0x4: {
				// 4XNN
				// Skips the next instruction if VX NOT equals NN
				let secondNibble = firstOpcode & 0xF; //0X00
				let val = opcode & 0x00FF; //0x00NN
				if(this.registers[secondNibble] != val){
					this.incrementPC();
				}
				this.incrementPC();
				}
				break;
			case 0x5: {
				// 5XY0
				// Skips the next instruction if VX == VY
				let secondNibble = firstOpcode & 0xF; //0X00
				let thirdNibble = secondOpcode >> 4; //00Y0
				if(this.registers[secondNibble] == this.registers[thirdNibble]) {
					this.incrementPC();
				}
				this.incrementPC();
				}
				break;
			case 0x6: {
				// 6XNN
				// Sets VX to NN
				let secondNibble = firstOpcode & 0xF; //0X00
				this.registers[secondNibble] = secondOpcode;
				this.incrementPC();
				}
				break;
			case 0x7: {
				// 7XNN
				// Adds NN to VX
				let secondNibble = firstOpcode & 0xF; //0X00
				this.registers[secondNibble] = secondOpcode;
				this.incrementPC();
				}
				break;
			case 0x8: {
				let secondNibble = firstOpcode & 0xF; //0X00
				let thirdNibble = secondOpcode >> 4; //00Y0
				let lastNibble = secondOpcode & 0xF; //000N
				if(lastNibble == 0x0) {
					// 8XY0
					// Vx |= Vy
					this.registers[secondNibble] = this.registers[thirdNibble];
					this.incrementPC();
				}else if(lastNibble == 0x1) {
					// 8XY1
					// Vx |= Vy
					this.registers[secondNibble] |= this.registers[thirdNibble];
					this.incrementPC();
				}else if(lastNibble == 0x2) {
					// 8XY2
					// Vx &= Vy
					this.registers[secondNibble] &= this.registers[thirdNibble];
					this.incrementPC();
				}else if(lastNibble == 0x3) {
					// 8XY3
					// Vx ^= Vy
					this.registers[secondNibble] ^= this.registers[thirdNibble];
					this.incrementPC();
				}else if(lastNibble == 0x4) {
					// 8XY4
					// Vx += Vy
					let sum = this.registers[secondNibble] + this.registers[thirdNibble];
					this.registers[0xF] = sum > 255 ? 1 : 0; // Set carry flag
					this.registers[secondNibble] = sum & 0xFF; // Get only the first byte
					this.incrementPC();
				}else if(lastNibble == 0x5) {
					// 8XY5
					// Vx -= Vy
					this.registers[0xF] = this.registers[secondNibble] > this.registers[thirdNibble] ? 1 : 0;
					let subs = this.registers[secondNibble] - this.registers[thirdNibble];
					this.registers[secondNibble] = subs;
					this.incrementPC();
				}else if(lastNibble == 0x6) {
					// 8XY6
					// Vx >>= 1
					this.registers[0xF] = this.registers[0xF] & 1; // ???? // Store the least significant bit of VX prior
					this.registers[secondNibble] >>= 1;
					this.incrementPC();
				}else if(lastNibble == 0x7) {
					// 8XY7
					// Vx = Vy - Vx
					this.registers[0xF] = this.registers[thirdNibble] > this.registers[secondNibble] ? 1 : 0;
					let subs = this.registers[thirdNibble] - this.registers[secondNibble];
					this.registers[secondNibble] = subs;
					this.incrementPC();
				}else if(lastNibble == 0xE) {
					// 8XYE
					// Vx <<= 1
					// Set VF to 1 if the most significant bit of VX prior to that shift was set
					this.registers[0xF] = this.registers[0xF] >> 3;
					this.registers[secondNibble] <<= 1;
					this.incrementPC();
				}

				}
				break;
			case 0x9: {
				// 9XY0
				// Skips the next instruction if VX != VY
				let secondNibble = firstOpcode & 0xF; //0X00
				let thirdNibble = secondOpcode >> 4; //00Y0
				if(this.registers[secondNibble] != this.registers[thirdNibble]) {
					this.incrementPC();
				}
				this.incrementPC();
				}
				break;
			
			case 0xA: {
				// ANNN
				// Sets I to the address NNN
				this.indexReg = opcode & 0x0FFF;
				this.incrementPC();
				}
				break;
			case 0xB:{
				// BNNN
				// Jumps to the address NNN plus V0
				this.pc = this.registers[0] + opcode & 0x0FFF;
				}
				break;
			case 0xC: {
				// CXNN
				// Vx = rand() & NN
				let secondNibble = firstOpcode & 0xF; //0X00
				this.registers[secondNibble] = this.rng.next255() & secondOpcode;
				this.incrementPC();
				}
				break;
			case 0xD: {
				// DXYN
				// draw(Vx, Vy, N)
				let secondNibble = firstOpcode & 0xF; //0X00
				let thirdNibble = secondOpcode >> 4; //00Y0
				let fourthNibble = secondOpcode & 0xF;
				
				this.registers[0xF] = 0;

				let regX = this.registers[secondNibble];
				let regY = this.registers[thirdNibble];
				
				let y = 0;
				while(y < fourthNibble) {
					// A sprite is always 8 bits horizontal
					let pixel = this.memory[this.indexReg + y];
					let x = 0;
					while(x < 8) {
						const msb = 0x80;
						

						x++;
					}

					y++;
				}

				this.incrementPC();
				}
				break;
		}
	}

	
}

export default CPU;